package Mentat::Message::Query::Mongo;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Message::Query::Mongo - Short module description

=head1 SYNOPSIS

 use Mentat::Message::Query::Mongo;

 # Instantinate the message query object
 my $query = Mentat::Message::Query::Mongo->new(Data::Convertor::Mongo->new(), Mentat::Storage::Mongo->new());

 # Query the MongoDB database for messages that pass the given filter
 my $messages = $query->query($filter);

=head1 DESCRIPTION



=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Data::Query::Mongo(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use JSON;

#-- Custom application modules ------------------------------------------------#
use Data::Query::Mongo;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.6';
    $DEVEL   = 0;
    @ISA     = ('Data::Query::Mongo');
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item query_msg($;$$$) [PUBLIC]

  Usage     : my $mentat_messages = $instance->query_msg($filter, $sort, $limit, $skip, $quick);
  Purpose   : Query the MongoDB database for Mentat messages
  Arguments : Data::Filter $filter - query filter tree
              HASH REFERENCE $sort - sorting specification
              INTEGER $limit - maximal number of records returned
              INTEGER $skip - number of records to skip from beginning
  Returns   : ARRAY REFERENCE containing Mentat::Message objects
  Throws    : Croaks, if not invoked on object

=cut

sub query_msg($;$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $sort, $limit, $skip, $quick) = @_;
    
    # Build the query hash
    my $query = $self->generate_query($filter);
    
    # Query the database
    return $self->storage->find_msg($query, $sort, $limit, $skip, $quick);
}

=item query_msg_wm($;$$$) [PUBLIC]

  Usage     : my $mentat_messages = $instance->query_msg_wm($filter, $sort, $limit, $skip, $quick);
  Purpose   : Query the MongoDB database for Mentat messages
  Arguments : Data::Filter $filter - query filter tree
              HASH REFERENCE $sort - sorting specification
              INTEGER $limit - maximal number of records returned
              INTEGER $skip - number of records to skip from beginning
  Returns   : ARRAY REFERENCE of HASHES
  Throws    : Croaks, if not invoked on object

=cut

sub query_msg_wm($;$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $sort, $limit, $skip, $quick) = @_;
    
    # Build the query hash
    my $query = $self->generate_query($filter);
    
    # Query the database
    return $self->storage->find_msg_wr($query, $sort, $limit, $skip, $quick);
}

=item query_msg_json($;$$$) [PUBLIC]

  Usage     : my $mentat_messages = $instance->query_msg_json($filter, $sort, $limit, $skip)
  Purpose   : Query the MongoDB database for Mentat messages in JSON
  Arguments : JSON STRING $filter - query filter tree
              JSON STRING $sort - sorting specification
              INTEGER $limit - maximal number of records returned
              INTEGER $skip - number of records to skip from beginning
  Returns   : ARRAY REFERENCE containing Mentat::Message objects
  Throws    : Croaks, if not invoked on object

=cut

sub query_msg_json($;$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $sort, $limit, $skip, $quick) = @_;
    
    my ($flt_j, $srt_j);
    if (ref($filter) eq 'HASH')    { $flt_j = $filter; }
    else                           { $flt_j = ($filter) ? decode_json($filter) : undef; }
    if (ref($sort) eq 'HASH')      { $srt_j = $sort; }
    else                           { $srt_j = ($sort) ? decode_json($sort) : undef; }
    
    # Query the database
    return $self->storage->find_msg($flt_j, $srt_j, $limit, $skip, $quick);
}

=item query_msg_json_wm($;$$$) [PUBLIC]

  Usage     : my $mentat_messages = $instance->query_msg_json_wm($filter, $sort, $limit, $skip)
  Purpose   : Query the MongoDB database for Mentat messages in JSON
  Arguments : JSON STRING $filter - query filter tree
              JSON STRING $sort - sorting specification
              INTEGER $limit - maximal number of records returned
              INTEGER $skip - number of records to skip from beginning
  Returns   : ARRAY REFERENCE containing Mentat::Message objects
  Throws    : Croaks, if not invoked on object

=cut

sub query_msg_json_wm($;$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $sort, $limit, $skip, $quick) = @_;
    
    my ($flt_j, $srt_j);
    if (ref($filter) eq 'HASH')    { $flt_j = $filter; }
    else                           { $flt_j = ($filter) ? decode_json($filter) : undef; }
    if (ref($sort) eq 'HASH')      { $srt_j = $sort; }
    else                           { $srt_j = ($sort) ? decode_json($sort) : undef; }
    
    # Query the database
    return $self->storage->find_msg_wr($flt_j, $srt_j, $limit, $skip, $quick);
}

=pod

=back

=cut

1;
