package Mentat::Reporter::Engine::Default;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Reporter::Engine::Default

=head1 SYNOPSIS

 use Mentat::Reporter::Engine::Default;

=head1 DESCRIPTION



=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Mentat::Reporter::Engine(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Storable 'dclone';
use IO::Compress::Zip qw(zip $ZipError);
use Text::Wrap;
use Number::Format qw(:subs :vars);
use Cwd 'abs_path';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Value::Convertor;
use Mentat::Core qw(:c_general :c_hawat);
use Mentat::Reporter::Engine;
use Mentat::Storage::Mongo;
use Data::Filter;
use Mentat::Message::Value::IDEA;
use Mentat::Reporter::Fetcher::Default;

use Mentat::Message::Grouper;
use Mentat::Stats::Data;
use Mentat::Message::Formater::TextTemplate;
use Mentat::Mailer::Report;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant DFLT_DATABASE          => 'mentat';
use constant DFLT_REPORT_COLLECTION => 'reports_ng';
use constant DFLT_GROUP_COLLECTION  => 'groups';

use constant DFLT_MODE     => 'summary';
use constant DFLT_TEMPLATE => 'default';
use constant DFLT_ATTACH   => 'all';
use constant DFLT_REDIRECT => 0;
use constant DFLT_MUTE     => 0;
use constant DFLT_COMPRESS => 1;

use constant TYPE_SUMMARY => Mentat::Message::Formater::TextTemplate::TYPE_SUMMARY;
use constant TYPE_EXTRA   => Mentat::Message::Formater::TextTemplate::TYPE_EXTRA;

use constant MODE_BOTH => 'both';

use constant ABUSE_UNKNOWN => '__UNKNOWN__';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Valid reporting flags
my $VALID_FLAGS = {
        'mute'     => 1,
        'redirect' => 1,
        'compress' => 1,
    };

# Valid reporting modes
my $VALID_MODES = {
        TYPE_SUMMARY() => 1,
        TYPE_EXTRA()   => 1,
        MODE_BOTH()    => 1,
    };

my $DEFAULT_CONFIGURATIONS = {
        filter   => undef,
        redirect => DFLT_REDIRECT(),
        mute     => DFLT_MUTE(),
        compress => DFLT_COMPRESS(),
        mode     => DFLT_MODE(),
        template => DFLT_TEMPLATE(),
        attach   => DFLT_ATTACH(),
        timing   => {
            'low'      => {period => '1day',   threshold => '6days', relapse => '2days'},
            'medium'   => {period => '2hrs',   threshold => '2days', relapse => '1day'},
            'high'     => {period => '10mins', threshold => '2hrs',  relapse => 'none'},
            'critical' => {period => '10mins', threshold => 'none',  relapse => 'none'},
        },
    };

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Mentat::Reporter::Engine');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item process(%) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub process(%)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my %config = @_;

    $config{time}         = time() unless $config{time};
    $config{last_run_log} = {}     unless $config{last_run_log};
    $config{groups}       = []     unless $config{groups};
    $config{groups}       = [$config{groups}] unless ref $config{groups} eq 'ARRAY';

    my $prefetch_period = ($self->{TEST_DATA}) ? '4hrs' : '6days';
    my $fetcher = $self->_fetcher($config{time}, $prefetch_period);

    $config{groups} = $fetcher->fetch_abuse_list() unless @{$config{groups}};

    my $ctime  = time;
    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Prefetch period: '$prefetch_period', processing basetime '".(scalar localtime($config{time}))." = '".$config{time}."')");

    # Prepare storage for report processing statistics
    my $process_run_log = {
            'ts_process' => $ctime,             # Timestamp of processing start
            'severities' => {},                 # Storage for particular severities
            'test_data'  => $self->{TEST_DATA}, # Flag for test runs
        };

    # Storage for list of generated reports
    my $reports = [];

    my ($mode, $timing, $sev_timing, $sev_duration, $last_timing, $tl, $th, $messages, $msg_count, $r);
    foreach my $group (@{$config{groups}}) {
        $mode   = $self->_get_group_config($group, 'mode');
        $timing = $self->_get_group_config($group, 'timing');

        # Override custom timing for test data mode
        if ($self->{TEST_DATA}) {
            Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Running is test data mode, setting period to '1hr' and threshold and relapse to 'none'");
            $timing = {
                'low'      => { 'period' =>  '1hr', 'threshold' =>  'none', 'relapse' =>  'none' },
                'medium'   => { 'period' =>  '1hr', 'threshold' =>  'none', 'relapse' =>  'none' },
                'high'     => { 'period' =>  '1hr', 'threshold' =>  'none', 'relapse' =>  'none' },
                'critical' => { 'period' =>  '1hr', 'threshold' =>  'none', 'relapse' =>  'none' },
            };
        }

        foreach my $severity ('critical', 'high', 'medium', 'low') {
            # Calculate the period boundaries
            $sev_timing   = $timing->{$severity};
            $sev_duration = $self->time_period($sev_timing->{period});
            $last_timing  = ($config{last_run_log} and ref $config{last_run_log} eq 'HASH' and $config{last_run_log}->{'last_ts_'.$severity}) ? $config{last_run_log}->{'last_ts_'.$severity} : undef;
            ($tl, $th)    = $self->calculate_time_boundaries(period => $sev_timing->{period}, time => $config{time}, offset => 1);
            Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Last processing of '$severity' was at '".(($last_timing) ? ((scalar localtime($last_timing))." = '".$last_timing) : "never"));

            unless ($config{skip_timing_check}) {
                # Check, that we are not running too soon
                unless ($self->check_timing($sev_timing->{period}, $config{time}, $last_timing)) {
                    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Skipping processing of severity '$severity', too soon (last run at '".(scalar localtime($last_timing))." = '".$last_timing."', required period '$sev_timing->{period}' period");
                    $process_run_log->{severities}->{$severity} = {'ts_last' => $last_timing, 'period' => $sev_timing->{period}, 'ts_next' => ($last_timing + $sev_duration)};
                    next;
                }
                else {
                    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Last processing of severity '$severity' occured at '".(($last_timing) ? ((scalar localtime($last_timing))." = '".$last_timing) : "never")."', required period '$sev_timing->{period}' period matched");
                    $process_run_log->{severities}->{$severity} = {'ts_last' => $last_timing, 'period' => $sev_timing->{period}, 'ts_from' => $tl, 'ts_to' => $th};
                    if ($last_timing and $last_timing < $tl) {
                        Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Ajusting lower boundary from '".(scalar localtime($tl))." = '".$tl."' to '".(scalar localtime($last_timing))." = '".$last_timing."'");
                        $tl = $last_timing;
                        $process_run_log->{severities}->{$severity}->{ts_from} = $tl;
                    }
                }
            }
            else {
                Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Skipping the timing check based on the configuration");
            }

            # Fetch the messages
            $messages  = $self->_fetch_messages($fetcher, $tl, $th, $group, $severity);
            $msg_count = scalar @$messages;

            # Skip the processing stage, when there is nothing to do
            next unless ($msg_count);

            Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: *** Starting the processing '$msg_count' messages of '$severity' severity with '$sev_timing->{period}' period");
            $r = $self->subprocess(messages => $messages,
                                   group => $group,
                                   severity => $severity,
                                   threshold => $sev_timing->{threshold},
                                   relapse => $sev_timing->{relapse},
                                   tl => $tl,
                                   th => $th,
                                   mode => $mode,
                                   process_run_log => $process_run_log);
            push (@$reports, @$r) if @$r;
            Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: *** Done processing messages of '$severity' severity");
        }
        Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: *** Done processing");
    }
    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] *** Done processing");

    my $fstatus = $fetcher->get_status();

    $process_run_log->{ts_done}            = time;
    $process_run_log->{duration}           = ($process_run_log->{ts_done} - $process_run_log->{ts_process});
    $process_run_log->{events_all}         = $fstatus->{counts}->{all};
    $process_run_log->{events_flt}         = $fstatus->{counts}->{flt};
    $process_run_log->{events_blk}         = $fstatus->{counts}->{blk};
    $process_run_log->{duration_fetch}     = $fstatus->{durations}->{fetch};
    $process_run_log->{duration_prefilter} = $fstatus->{durations}->{prefilter};

    # Calculate the fetching and prefiltering speed
    $process_run_log->{speed_fetch}     = ($process_run_log->{duration_fetch})     ? ($process_run_log->{events_all} / $process_run_log->{duration_fetch})     : 0;
    $process_run_log->{speed_prefilter} = ($process_run_log->{duration_prefilter}) ? ($process_run_log->{events_all} / $process_run_log->{duration_prefilter}) : 0;

    return ($process_run_log, $reports);
}

=item process_legacy(%) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub process_legacy(%)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my %config = @_;

    $config{time}         = time() unless $config{time};
    $config{last_run_log} = {}     unless $config{last_run_log};
    $config{groups}       = []     unless $config{groups};

    my $fetcher = $self->_fetcher($config{time}, '4hrs');

    $config{groups} = $fetcher->fetch_abuse_list() unless @{$config{groups}};

    my $ctime  = time;
    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] LEGACY: Processing basetime '".(scalar localtime($config{time}))." = '".$config{time}."')");

    # Prepare storage for report processing statistics
    my $process_run_log = {
            'ts_process' => $ctime,             # Timestamp of processing start
            'test_data'  => $self->{TEST_DATA}, # Flag for test runs
        };

    # Storage for list of generated reports
    my $reports = [];

    my ($mode, $period, $duration, $last_timing, $tl, $th, $messages, $msg_count, $r);
    foreach my $group (@{$config{groups}}) {
        $mode   = $self->_get_group_config($group, 'mode');
        $period = '2hrs';
        $duration = $self->time_period($period);
        $last_timing  = ($config{last_run_log} and ref $config{last_run_log} eq 'HASH' and $config{last_run_log}->{'last_ts_legacy'}) ? $config{last_run_log}->{'last_ts_legacy'} : undef;
        ($tl, $th)    = $self->calculate_time_boundaries(period => $period, time => $config{time}, offset => 1);

        unless ($config{skip_timing_check}) {
            # Check, that we are not running too soon
            unless ($self->check_timing($period, $config{time}, $last_timing)) {
                Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] LEGACY: Skipping processing '$group', too soon to do anything (last run at '".(scalar localtime($last_timing))." = '".$last_timing."', required period '$period' period");
                $process_run_log->{legacy} = {'ts_last' => $last_timing, 'period' => $period, 'ts_next' => ($last_timing + $duration)};
                next;
            }
            else {
                Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] LEGACY: Last processing '$group' occured at '".(($last_timing) ? ((scalar localtime($last_timing))." = '".$last_timing) : "never")."', required period '$period' period matched");
                $process_run_log->{legacy} = {'ts_last' => $last_timing, 'period' => $period, 'ts_from' => $tl, 'ts_to' => $th};
                if ($last_timing and $last_timing < $tl) {
                    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] LEGACY: Ajusting lower boundary from '".(scalar localtime($tl))." = '".$tl."' to '".(scalar localtime($last_timing))." = '".$last_timing."'");
                    $tl = $last_timing;
                    $process_run_log->{legacy}->{ts_from} = $tl;
                }
            }
        }
        else {
            Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] LEGACY: Skipping the timing check based on the configuration");
        }

        # Fetch the messages
        $messages  = $self->_fetch_abuse_messages($fetcher, $tl, $th, $group);
        $msg_count = scalar @$messages;

        # Skip the processing stage, when there is nothing to do
        next unless ($msg_count);

        Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] *** LEGACY: Starting the processing '$msg_count' messages for '$group' group with '$period' period");
        $r = $self->subprocess(messages => $messages,
                               group => $group,
                               tl => $tl,
                               th => $th,
                               skip_thresholding => 1,
                               skip_relapse => 1,
                               mode => $mode,
                               process_run_log => $process_run_log);
        push (@$reports, @$r) if @$r;
        Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] *** LEGACY: Done processing '$group'");
    }
    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] *** LEGACY: Done processing");

    my $fstatus = $fetcher->get_status();

    $process_run_log->{ts_done}            = time;
    $process_run_log->{duration}           = ($process_run_log->{ts_done} - $process_run_log->{ts_process});
    $process_run_log->{events_all}         = $fstatus->{counts}->{all};
    $process_run_log->{events_flt}         = $fstatus->{counts}->{flt};
    $process_run_log->{events_blk}         = $fstatus->{counts}->{blk};
    $process_run_log->{duration_fetch}     = $fstatus->{durations}->{fetch};
    $process_run_log->{duration_prefilter} = $fstatus->{durations}->{prefilter};

    # Calculate the fetching and prefiltering speed
    $process_run_log->{speed_fetch}     = ($process_run_log->{duration_fetch})     ? ($process_run_log->{events_all} / $process_run_log->{duration_fetch})     : 0;
    $process_run_log->{speed_prefilter} = ($process_run_log->{duration_prefilter}) ? ($process_run_log->{events_all} / $process_run_log->{duration_prefilter}) : 0;

    return ($process_run_log, $reports);
}

=item subprocess(%) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub subprocess(%)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my %config = @_;
    die "[ENG:Reporter:Engine] Messages must be specified under configuration key 'messages'\n" unless exists $config{messages};
    die "[ENG:Reporter:Engine] Abuse group must be specified under configuration key 'group'\n" unless exists $config{group};

    my $messages  = $config{messages};
    my $msg_count = scalar @{$config{messages}};

    my ($report, $r, $frv, $relapses, $msg_count_thr, $msg_count_rlp, $msg_count_flt);
    my $reports = [];

    # Apply the filtering
    unless ($self->{TEST_DATA}) {
        ($messages, $frv) = $self->_filter_messages($config{group}, $messages);
        $msg_count_flt = scalar @$messages;
        unless ($msg_count_flt) {
            Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$config{group}>: Filters blocked all '$msg_count' messages");
            return $reports;
        }
        Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$config{group}>: Filters blocked ".($msg_count - $msg_count_flt)." messages ($msg_count_flt passed)");
    }
    else {
        Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$config{group}>: Running in test data mode, skipping the filtering mechanics");
        $frv = {};
        $msg_count_flt = scalar @$messages;
    }

    # Apply the thresholding, when not disabled globally by configuration
    unless ($config{skip_thresholding}) {
        # Thresholding might be disabled by setting the threshold interval to 'none'
        unless ($config{threshold} eq 'none') {
            $messages = $self->_threshold_messages($config{group}, $messages, $config{threshold});
            $msg_count_thr = scalar @$messages;
            unless ($msg_count_thr) {
                Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$config{group}>: Thresholds blocked all '$msg_count_flt' messages");
                return $reports;
            }
            Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$config{group}>: Thresholds blocked ".($msg_count_flt - $msg_count_thr)." messages ($msg_count_thr passed)");
        }
        else {
            Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$config{group}>: Threshold interval set to 'none', skipping the thresholding mechanics");
            $msg_count_thr = scalar @$messages;
        }
    }
    else {
        Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$config{group}>: Skipping the thresholding mechanics");
        $msg_count_thr = scalar @$messages;
    }

    # Apply the relapse heuristics
    unless ($config{skip_relapse}) {
        unless ($config{relapse} eq 'none') {
            $relapses = $self->_relapse_messages($config{fetcher}, $config{severity}, $config{group}, $config{threshold}, $config{relapse});
            $msg_count_rlp = scalar @$relapses;
            if ($msg_count_rlp) {
                Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$config{group}>: Included '$msg_count_rlp' relapsed messages");
            }
        }
        else {
            Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$config{group}>: Relapse interval set to 'none', skipping the relapse mechanics");
            $msg_count_rlp = 0;
        }
    }
    else {
        Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$config{group}>: Skipping the relapse mechanics");
        $msg_count_rlp = 0;
    }

    # Prepare the report structure
    $report = {
            'abuse'       => $config{group},
            'severity'    => $config{severity},
            'ts_from'     => int($config{tl}),
            'ts_to'       => int($config{th}),
            'type'        => TYPE_SUMMARY(),
            'cnt_all'     => $msg_count,
            'cnt_flt'     => $msg_count_flt,
            'cnt_flt_blk' => ($msg_count - $msg_count_flt),
            'cnt_thr'     => $msg_count_thr,
            'cnt_thr_blk' => ($msg_count_flt - $msg_count_thr),
            'cnt_rlp'     => $msg_count_rlp,
            'frv'         => [ map { [$_, $frv->{$_}] } keys %$frv ],
            'test_data'   => $self->{TEST_DATA},
        };
    # Generate summary reports
    $r = $self->generate_report($messages, $relapses, $report, $config{process_run_log});
    push (@$reports, $r);

    # Generate individual reports
    if ($config{mode} eq TYPE_EXTRA() or $config{mode} eq MODE_BOTH()) {
        # Prepare the report structure
        $report = {
            'abuse'     => $config{group},
            'severity'  => $config{severity},
            'ts_from'   => int($config{tl}),
            'ts_to'     => int($config{th}),
            'type'      => TYPE_EXTRA(),
            'test_data' => $self->{TEST_DATA},
        };
        $r = $self->_process_abuse_ipextra($messages, $relapses, $report, $config{process_run_log});
        push (@$reports, @$r);
    }

    # Setup thresholding cache for reported problems
    unless ($config{skip_thresholding}) {
        unless ($config{threshold} eq 'none') {
            $self->_setup_thresholding_cache($messages);
        }
    }

    # Setup relapse cache for reported problems
    unless ($config{skip_relapse}) {
        unless ($config{relapse} eq 'none') {
            $self->_setup_relapse_cache($relapses);
        }
    }

    return $reports;
}

=item generate_report($$$;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub generate_report($$$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($messages, $relapses, $report, $process_run_log) = @_;

    my $template = $self->_get_group_config($report->{abuse}, 'template');

    # Generate unique report ID
    my $id = $self->_generate_report_id($report);
    $report->{id} = $id;
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$report->{abuse}>: Generating 'summary' report '$id'");

    # Generate unique hash for unauthorized access
    $report->{ua_hash} = $self->_generate_report_uahash($id);

    # Calculate statistics for messages within the report
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter] Report '$id': Calculating message statistics");
    $report = $self->{STATS}->evaluate_alerts($messages, $report);

    # Generate the report body
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter] Report '$id': Generating report message body");
    $report->{message} = $self->{FORMATER_TEXT}->format_report($id, $messages, $report, $template) unless $self->{SIMULATE};

    # Generate the report attachments
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter] Report '$id': Generating data attachment in JSON");
    $report->{attachment_json} = $self->{FORMATER_JSON}->format_report($id, $messages, $report) unless $self->{SIMULATE};
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter] Report '$id': Generating data attachment in CSV");
    $report->{attachment_csv}  = $self->{FORMATER_CSV}->format_report($id, $messages, $report)  unless $self->{SIMULATE};

    # Compress and store the report attachments
    $self->_report_store_attachment_json($id, $report) unless $self->{SIMULATE};
    $self->_report_store_attachment_csv( $id, $report) unless $self->{SIMULATE};

    $self->_report_email($id, $report)     unless $self->{SIMULATE};
    $self->_report_store_log($id, $report) unless $self->{SIMULATE};

    # Update run log statistics
    $self->_runlog_update($process_run_log, $report, $id) if $process_run_log;

    return $report;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Reporter::Engine::Default instance
#  Arguments : Unknown see the description in appropriate subclass
#  Returns   : Mentat::Reporter::Engine::Default $self reference
#  Throws    : Dies, if not invoked on instance or given invalid arguments
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Initializing Mentat::Reporter::Engine::Default instance...
    my %config = @_;

    die "[ENG:Reporter:Engine] Filter grammar file must be specified under configuration key 'filter_grammar_file'\n" unless exists $config{filter_grammar_file};
    die "[ENG:Reporter:Engine] Filter grammar file '$config{filter_grammar_file}' must be ordinary file\n" unless -f $config{filter_grammar_file};
    die "[ENG:Reporter:Engine] Filter schema file must be specified under configuration key 'filter_schema_file'\n" unless $config{filter_schema_file};
    die "[ENG:Reporter:Engine] Filter schema file '$config{filter_schema_file}' must be ordinary file\n" unless -f $config{filter_schema_file};
    die "[ENG:Reporter:Engine] Prefilter engine configuration 'filter_config' is missing\n" unless $config{filter_config};
    die "[ENG:Reporter:Engine] Prefilter engine configuration 'filter_config_test' is missing\n" unless $config{filter_config_test};
    die "[ENG:Reporter:Engine] Template directory must be specified under configuration key 'template_dir'\n" unless exists $config{template_dir};
    die "[ENG:Reporter:Engine] Template directory '$config{'template_dir'}' does not exist\n" unless -d $config{template_dir};
    die "[ENG:Reporter:Engine] Temp folder must be specified under configuration key 'temp_folder'\n" unless exists $config{temp_folder};
    die "[ENG:Reporter:Engine] Temp folder '$config{'temp_folder'}' does not exist\n" unless -d $config{temp_folder};
    die "[ENG:Reporter:Engine] Temp folder '$config{'temp_folder'}' is not writable\n" unless -w $config{temp_folder};
    die "[ENG:Reporter:Engine] Chart directory must be specified under configuration key 'charts_dir'\n" unless exists $config{charts_dir};
    die "[ENG:Reporter:Engine] Chart directory '$config{charts_dir}' does not exist\n" unless -d $config{charts_dir};
    die "[ENG:Reporter:Engine] Chart directory '$config{charts_dir}' is not writable\n" unless -w $config{charts_dir};
    die "[ENG:Reporter:Engine] Report directory must be specified under configuration key 'reports_dir'\n" unless exists $config{reports_dir};
    die "[ENG:Reporter:Engine] Report directory '$config{'reports_dir'}' does not exist\n" unless -d $config{reports_dir};
    die "[ENG:Reporter:Engine] Report directory '$config{'reports_dir'}' is not writable\n" unless -w $config{reports_dir};
    die "[ENG:Reporter:Engine] Name of the event database must be specified under configuration key 'event_database'\n" unless $config{event_database};
    die "[ENG:Reporter:Engine] Name of the event collection must be specified under configuration key 'event_collection'\n" unless $config{event_collection};

    $self->{RSTORAGE} = Mentat::Storage::Mongo->new(database => DFLT_DATABASE(), collection => DFLT_REPORT_COLLECTION());
    $self->{GSTORAGE} = Mentat::Storage::Mongo->new(database => DFLT_DATABASE(), collection => DFLT_GROUP_COLLECTION());

    $self->{CACHE} = $config{cache};
    $self->_cleanup_cache();

    $self->{FILTER_GRAMMAR}     = $config{filter_grammar_file};
    $self->{FILTER_SCHEMA}      = $config{filter_schema_file};
    $self->{FILTER_CONFIG}      = $config{filter_config};
    $self->{FILTER_CONFIG_TEST} = $config{filter_config_test};
    $self->{SIMULATE}           = $config{simulate};
    $self->{TEST_DATA}          = ($config{test_data}) ? 1 : 0;

    $self->{GROUPER} = Mentat::Message::Grouper->new();
    $self->{STATS}   = Mentat::Stats::Data->new(%config);
    $self->{HELPER}  = Mentat::Message::Formater::Helper->service();
    $self->{FORMATER_TEXT} = Mentat::Message::Formater::TextTemplate->new($config{template_dir} . '/email',
                                                                          $self->{HELPER});
    $self->{FORMATER_JSON} = Mentat::Message::Formater::TextTemplate->new($config{template_dir} . '/json',
                                                                          $self->{HELPER});
    $self->{FORMATER_CSV}  = Mentat::Message::Formater::TextTemplate->new($config{template_dir} . '/csv',
                                                                          $self->{HELPER});

    $self->{MAILER}      = Mentat::Mailer::Report->new(%config);
    $self->{TEMP_FOLDER} = abs_path($config{temp_folder});
    $self->{REPORTS_DIR} = abs_path($config{reports_dir});

    $self->_load_group_configs();

    # Prepare storage for report processing statistics
    $self->{RUNLOG} = {
            'ts_process' => time, # Timestamp of processing start
        };

    # Ensure the storage indexes
    $self->{RSTORAGE}->_create_indexes([{'ts_from'  => 1}, {'background' => 1}],
                                       [{'ts_to'    => 1}, {'background' => 1}],
                                       [{'ts'       => 1}, {'background' => 1}],
                                       [{'abuse'    => 1}, {'background' => 1}],
                                       [{'list_ips' => 1}, {'background' => 1}],
        );

    $self->{TOGGLE_EMAIL} = (exists $config{toggle_email}) ? $config{toggle_email} : 1;
    $self->{TOGGLE_MONGO} = (exists $config{toggle_mongo}) ? $config{toggle_mongo} : 1;

    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Running in simulation mode, no changes will be made in DB or on disk, no reports will be mailed") if $self->{SIMULATE};
    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Running in test data mode, only 'Test' category data will be processed") if $self->{TEST_DATA};

    return $self;
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _fetcher($$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($ref_time, $period, $abuse) = @_;

    my ($tl, $th) = $self->calculate_time_boundaries(period => $period, time => $ref_time, noadjust => 1);

    my $fc = $self->{FILTER_CONFIG};
    if ($self->{TEST_DATA}) {
        $fc = $self->{FILTER_CONFIG_TEST};
    }

    return Mentat::Reporter::Fetcher::Default->new(filter_grammar_file => $self->{FILTER_GRAMMAR},
                                                   filter_schema_file  => $self->{FILTER_SCHEMA},
                                                   filter_config       => $fc,
                                                   tl                  => $tl,
                                                   th                  => $th,
                                                   abuse               => $abuse);
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _load_group_configs()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my ($groups, $count_all) = $self->{GSTORAGE}->find();
    unless ($count_all) {
        Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Database does not contain any group reporting configurations");
        $self->{CONFIGURATIONS} = {};
        return;
    }

    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Loading all group reporting configurations from database");
    my $gid;
    foreach my $group (@$groups) {
        $gid = $group->{_id};
        die "Configuration for abuse group '$gid' was already defined\n" if exists $self->{CONFIGURATIONS}->{$gid};
        my $act = {}; $self->{CONFIGURATIONS}->{$gid} = $act;

        $act->{email} = (exists($group->{rep_emails}) and $group->{rep_emails}) ? $group->{rep_emails} : $gid;

        $act->{redirect} = (exists($group->{rep_redirect})) ? $group->{rep_redirect} : DFLT_REDIRECT();
        $act->{mute}     = (exists($group->{rep_mute}))     ? $group->{rep_mute}     : DFLT_MUTE();
        $act->{compress} = (exists($group->{rep_compress})) ? $group->{rep_compress} : DFLT_COMPRESS();
        $act->{mode}     = (exists($group->{rep_mode}))     ? $group->{rep_mode}     : DFLT_MODE();
        $act->{attach}   = (exists($group->{rep_attach}))   ? $group->{rep_attach}   : DFLT_ATTACH();
        $act->{template} = (exists($group->{rep_template})) ? $group->{rep_template} : DFLT_TEMPLATE();

        if (exists $group->{rep_filters}) {
            eval {
                ($act->{filter}, $act->{trtf}) = $self->_build_filter($group->{rep_filters}, $gid);
            };
            if ($@) {
                Log::Loger->error(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$gid>: Error in filter instantination: $@");
            }
        }
    }

    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Successfully loaded all group reporting configurations from database");
    Log::Loger->debug(__PACKAGE__, "CONFIGURATIONS: " . Dumper($self->{CONFIGURATIONS})) if $DEVEL;
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _get_group_config($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($group, $key) = @_;

    my $cfg;
    if (exists $self->{CONFIGURATIONS}->{$group} and exists $self->{CONFIGURATIONS}->{$group}->{$key} and defined $self->{CONFIGURATIONS}->{$group}->{$key}) {
        $cfg = $self->{CONFIGURATIONS}->{$group}->{$key};
        Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Using abuse specific '$key' value '$cfg'") unless ref $cfg;
        if ($key eq 'email' and not $cfg) {
            $cfg = $group;
            Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Substitution for empty mandatory key '$key' with value '$cfg'");
        }
    }
    else {
        if ($key eq 'email') { $cfg = $group; }
        else                 { $cfg = $DEFAULT_CONFIGURATIONS->{$key}; }
        unless (defined $cfg) {
            Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Using undefined default '$key' value");
        }
        elsif (not ref $cfg) {
            Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Using default '$key' value '$cfg'");
        }
    }
    return $cfg;
}

#-------------------------------------------------------------------------------

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _build_filter($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filters, $group) = @_;

    my $args = {
            'compile' => 'idea',
            'order' => 'deny',
            'deny' => [],
            'default_action' => 'allow',
        };
    my $trtf = {};
    my $i = 0;
    my $ts = time();
    foreach my $f (@$filters) {
        # Filter may be explicitly disabled by configuration, check for this case
        unless ($f->{enabled}) {
            Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Skipped disabled filter '$f->{_id}': \"$f->{filter}\"");
            next;
        }
        # Filter may have validity period
        if ($f->{validfrom} and int($f->{validfrom}) and int($f->{validfrom}) > $ts) {
            Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Skipped filter '$f->{_id}': \"$f->{filter}\", not valid before ".Value::Convertor->unixs_to_ldatestr(int($f->{validfrom})));
            next;
        }
        # Filter may have validity period
        if ($f->{validto} and int($f->{validto}) and int($f->{validto}) < $ts) {
            Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Skipped filter '$f->{_id}': \"$f->{filter}\", not valid after ".Value::Convertor->unixs_to_ldatestr(int($f->{validto})));
            next;
        }

        Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Loaded filter '$f->{_id}': \"$f->{filter}\"");

        # Insert the filter among the factory arguments
        push(@{$args->{deny}}, $f->{filter});
        # Insert the record into translation table
        $trtf->{++$i} = $f;
    }

    return (Data::Filter->new(filter_grammar_file => $self->{FILTER_GRAMMAR},
                              filter_schema_file => $self->{FILTER_SCHEMA},
                              filter_config => $args,
                              dt_resolver => Mentat::Message::Value::IDEA->new()),
            $trtf);
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _build_relapse_map($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($fetcher, $severity, $group, $relapse) = @_;

    # Calculate the period boundaries
    my ($tl, $th) = $self->calculate_time_boundaries(period => $relapse, offset => 1);

    my $map = {};
    my $possible_relapses       = $self->_fetch_messages($fetcher, $tl, $th, $group, $severity);
    my $possible_relapses_count = scalar @$possible_relapses;

    # Skip the processing stage, when there is nothing to do
    return $map unless ($possible_relapses_count);

    # Filter out the messages, that are not thresholded
    $possible_relapses = $self->_filter_thresholded_messages($group, $possible_relapses, $relapse);

    my @keys;
    foreach my $msg (@$possible_relapses) {
        @keys = $self->_msg_gen_keys($msg);
        foreach my $key (@keys) {
            $map->{$key}++;
        }
    }

    return $map;
}

#-------------------------------------------------------------------------------

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _fetch_messages($$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($fetcher, $tl, $th, $group, $severity) = @_;
    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Fetching messages with severity '$severity' ('".(scalar localtime($tl))." - ".(scalar localtime($th))."' = '$tl - $th')");

    # Handle the empty time period
    if ($tl == $th) {
        Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Nothing to fetch for empty period ('".(scalar localtime($tl))." - ".(scalar localtime($th))."' = '$tl - $th')");
        return [];
    }

    # Fetch the messages
    my $messages  = $fetcher->get_data($tl, $th, $group, $severity);
    my $msg_count = scalar @$messages;
    unless ($msg_count) {
        Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Fetched no messages with '$severity' severity ('".(scalar localtime($tl))." - ".(scalar localtime($th))."' = '$tl - $th')");
    }
    else {
        Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Fetched '$msg_count' messages with '$severity' severity ('".(scalar localtime($tl))." - ".(scalar localtime($th))."' = '$tl:$th')");
    }

    return $messages;
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _fetch_abuse_messages($$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($fetcher, $tl, $th, $group) = @_;
    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Fetching all messages ('".(scalar localtime($tl))." - ".(scalar localtime($th))."' = '$tl - $th')");

    # Handle the empty time period
    if ($tl == $th) {
        Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Nothing to fetch for empty period ('".(scalar localtime($tl))." - ".(scalar localtime($th))."' = '$tl - $th')");
        return [];
    }

    # Fetch the messages
    my $messages  = $fetcher->get_abuse_data($tl, $th, $group);
    my $msg_count = scalar @$messages;
    unless ($msg_count) {
        Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Fetched no messages ('".(scalar localtime($tl))." - ".(scalar localtime($th))."' = '$tl:$th')");
    }
    else {
        Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Fetched '$msg_count' messages ('".(scalar localtime($tl))." - ".(scalar localtime($th))."' = '$tl:$th')");
    }

    return $messages;
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _filter_messages($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($group, $messages) = @_;
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Filtering out messages");

    my $filter = $self->_get_group_config($group, 'filter');
    my $frv = {};
    if ($filter) {
        my $temp = [];

        my ($result, $rule, $idx);
        foreach my $msg (@$messages) {
            ($result, $rule, $idx) = $filter->filter($msg);
            if ($result) {
                push(@$temp, $msg);
            }
            else {
                $frv->{$idx.":".$rule}++;
            }
        }

        $messages = $temp;
    }

    $self->_log_filtering($group, $frv);
    return ($messages, $frv);
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _filter_expired_messages($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($group, $messages) = @_;
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Filtering out expired messages");

    my @result = ();
    foreach my $msg (@$messages) {
        # Include only messages, that are not expired
        push(@result, $msg) unless $msg->is_expired($self->time_period('2days'));
    }

    return \@result;
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _filter_thresholded_messages($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($group, $messages, $threshold) = @_;

    my $ttl = $self->time_period($threshold);
    my $tsl = time() - $ttl;
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Filtering out thresholded messages, using '$tsl' as threshold (".Value::Convertor->unixs_to_ldatestr($tsl).")");

    my @result = ();
    foreach my $msg (@$messages) {
        # Include only messages, that are currently thresholded
        push(@result, $msg) if $self->_msg_is_thresholded($msg, $tsl);
    }

    return \@result;
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _filter_thresholded_exp_messages($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($group, $messages, $threshold) = @_;
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Filtering out expired thresholded messages");

    my $ttl = $self->time_period($threshold);
    my $tsl = time() - $ttl;
    my @result = ();
    foreach my $msg (@$messages) {
        # Include only messages, that have expired thresholding
        push(@result, $msg) if $self->_msg_is_thresholded_exp($msg, $tsl);
    }

    return \@result;
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _filter_relapsed_messages($$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($fetcher, $severity, $group, $messages, $relapse) = @_;
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Filtering relapsed messages");

    my $relapse_map = $self->_build_relapse_map($fetcher, $severity, $group, $relapse);

    my @result = ();
    foreach my $msg (@$messages) {
        # Include only messages, that currently relapsed
        push(@result, $msg) if $self->_msg_is_relapsed($msg, $relapse_map);
    }

    #return \@result;
    return [];
}

#-------------------------------------------------------------------------------

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _threshold_messages($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($group, $messages, $threshold) = @_;

    my $ttl = $self->time_period($threshold);
    my $tsl = time() - $ttl;
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Thresholding messages, using '$tsl' as threshold (".Value::Convertor->unixs_to_ldatestr($tsl).", TTL: $ttl, THR: $threshold)");

    my @result = ();
    foreach my $msg (@$messages) {
        next if $self->_msg_is_thresholded($msg, $tsl);
        push(@result, $msg);
    }

    return \@result;
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _relapse_messages($$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($fetcher, $severity, $group, $threshold, $relapse) = @_;
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Relapsing messages");

    ################## TODO: Extensively test the relapse, currently bypassed
    return [];

    # Calculate the period boundaries
    my ($tl, $th) = $self->calculate_time_boundaries(period => $threshold, offset => 1);

    # Fetch the messages
    my $messages     = $self->_fetch_messages($fetcher, $tl, $th, $group, $severity);
    my $messages_cnt = scalar @$messages;

    # Skip the processing stage, when there is nothing to do
    return $messages unless ($messages_cnt);

    # Filter out the messages, that are not thresholded anymore
    $messages = $self->_filter_thresholded_exp_messages($group, $messages, $threshold);

    # Filter out messages, that are not relapsed
    $messages = $self->_filter_relapsed_messages($severity, $group, $messages, $relapse);

    return $messages;
}

#-------------------------------------------------------------------------------

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _process_abuse_ipextra($$$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($messages, $relapses, $report_orig, $process_run_log) = @_;

    # Group messages according to the IP before generating reports
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter] Abuse <".$report_orig->{abuse}.">: Pre-grouping messages for 'extra' reports");
    $messages = $self->{GROUPER}->group('Source/IP4', $messages);

    # Generate extra report for each IP address
    my ($msgs, $message, $attachment, $attachment2, $id, $ua_hash, $report);
    my $reports = [];
    foreach my $ip (sort keys(%$messages)) {
        # Create copy of original report object
        $report = dclone($report_orig);

        # Store the IP address within the report
        $report->{ip} = $ip;

        # Store the number of events within the report
        $report->{cnt_all} = scalar @{$messages->{$ip}};

        # Dispatch to regular reporting algorithm with smaller dataset
        $report = $self->generate_report($messages->{$ip}, $relapses, $report, $process_run_log);

        # Insert the report to the report heap
        push(@$reports, $report);
    }
    return $reports;
}

#-------------------------------------------------------------------------------

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _cleanup_cache()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    return $self->{CACHE}->cleanup($self->time_period('3weeks'));
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _setup_thresholding_cache($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($messages,) = @_;

    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Updating threshold cache");

    my @keys;
    my $ts = time;
    foreach my $msg (@$messages) {
        @keys = $self->_msg_gen_keys($msg);
        $self->{CACHE}->set_threshold_cache(\@keys, $ts);
    }

    return $messages;
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _setup_relapse_cache($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($messages,) = @_;

    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Updating relapse cache");

    return $messages;
}

#-------------------------------------------------------------------------------

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _log_filtering($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($group, $frv) = @_;
    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Filtering result: ". Dumper($frv));

    # Fetch filter configurations for this abuse
    my $trtf = $self->_get_group_config($group, 'trtf');
    return unless ($trtf);

    my ($idx, $fid, $filter, $fspec_a, $fspec_b);
    # Process the filtering results
    foreach my $f (keys %$frv) {
        # Fetch the metadata for appropriate filter
        ($idx, $fspec_a) = split(/:/, $f);
        $filter = $trtf->{$idx};
        unless ($filter) {
            Log::Loger->warn(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Something weird happened, filtering result contains record for unknown filter '$f'");
            return;
        }

        # Pick appropriate metadata
        $fid     = $filter->{_id};
        $fspec_b = $filter->{filter};

        Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Filter '$fid':'$fspec_b' blocked '".$frv->{$f}."' messages");

        # Update the counters within the database
        my $rv = $self->{GSTORAGE}->update(
            { HAWAT_ATTR_MID() => $group, 'rep_filters.'.HAWAT_ATTR_MID() => $fid},
            { '$inc' => { 'rep_filters.$.hits' => $frv->{$f} }},
        );
        unless (ref $rv and $rv->{ok} == 1 and $rv->{n} == 1) {
            Log::Loger->error(__PACKAGE__, "[ENG:Reporter:Engine] Abuse <$group>: Could not store filtering results into database: ". Dumper($rv));
        }
    }
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _msg_is_thresholded($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($msg, $tsl) = @_;

    my @keys = $self->_msg_gen_keys($msg);
    return $self->{CACHE}->check_threshold_cache(\@keys, $tsl);
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _msg_is_thresholded_exp($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($msg, $tsl) = @_;

    my @keys = $self->_msg_gen_keys($msg);
    return $self->{CACHE}->check_relapse_cache(\@keys, $tsl);
}

# _ [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _msg_is_relapsed($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($msg, $map) = @_;

    my $rv; my $x;
    my @keys = $self->_msg_gen_keys($msg);
    foreach my $key (@keys) {
        $x = (exists $map->{$key}) ? 1 : 0;
        $rv = (defined($rv)) ? (($x and $rv)?1:0) : ($x);
    }
    return $rv;
}

#-------------------------------------------------------------------------------

# _generate_report_id() [PROTECTED]
#
#  Usage     : my $id = $self->_generate_report_id();
#  Purpose   : Generate unique report ID
#  Arguments : None
#  Returns   : STRING $id
#  Throws    : Dies, if not invoked on object instance

sub _generate_report_id(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($report,) = @_;

    my $severity = ($report and $report->{severity}) ? $report->{severity} : 'unknown';
    my $s        = substr(ucfirst($severity), 0, 1);

    my $type = ($report and $report->{type}) ? $report->{type} : 'unknown';
    my $t    = substr(ucfirst($type), 0, 1);

    # Calculate current label
    my ($tm, $label);
    {
        use Time::localtime;
        $tm    = localtime;
        $label = sprintf("M%04d%02d%02d", ($tm->year) + 1900, ($tm->mon) + 1, $tm->mday);
    }

    # Attempt to block the ID within the database
    my $id;
    while (1) {
        $id = sprintf("%s%1s%1s-%5s", $label, $t, $s, Mentat::Core::uid(undef, 5));
        next unless $self->_block_report_id($id);
        last;
    }

    return $id;
}

# _generate_report_uahash($) [PROTECTED]
#
#  Usage     : my $id = $self->_generate_report_uahash($id);
#  Purpose   : Generate unique report ID
#  Arguments : None
#  Returns   : STRING $id
#  Throws    : Dies, if not invoked on object instance

sub _generate_report_uahash($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($report_id,) = @_;

    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Engine] Report *$report_id*: Generating random hash for unauth access'");
    my $ua_hash = sprintf("%s", Mentat::Core::uid(undef, 20));
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Engine] Report *$report_id*: Generated hash for unauth access: '$ua_hash'");
    return $ua_hash;
}

# _block_report_id($) [PROTECTED]
#
#  Usage     : my $rv = $self->_block_report_id($id);
#  Purpose   : Block given ID for new report within the database
#  Arguments : STRING $id - requested ID [MANDATORY]
#  Returns   : BOOL $rv, TRUE on SUCCESS, FALSE on FAILURE
#  Throws    : Dies, if not invoked on object instance

sub _block_report_id($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($report_id,) = @_;

    return 1 if $self->{SIMULATE};

    eval {
        $self->{RSTORAGE}->insert({'_id' => $report_id});
    };
    if ($@) {
        Log::Loger->warn(__PACKAGE__, "[ENG:Reporter] Report with ID '$report_id' already exists in database");
        return 0;
    }
    Log::Loger->warn(__PACKAGE__, "[ENG:Reporter] Blocked unique ID '$report_id' for new report within database");
    return 1;
}

# _generate_report_attach_id($$;$) [PROTECTED]
#
#  Usage     : my $id = $self->_generate_report_attach_id($id, $type, $compression);
#  Purpose   :
#  Arguments : None
#  Returns   : STRING $id
#  Throws    : Dies, if not invoked on object instance

sub _generate_report_attach_id($$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($report_id, $type, $compression) = @_;

    my $suffix = ($compression)?'.zip':'';
    return "security_report_$report_id.$type".$suffix;
}

# _generate_report_json_id($;$) [PROTECTED]
#
#  Usage     : my $id = $self->_generate_report_json_id($id, $compression);
#  Purpose   :
#  Arguments : None
#  Returns   : STRING $id
#  Throws    : Dies, if not invoked on object instance

sub _generate_report_json_id($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($report_id, $compression) = @_;

    my $suffix = ($compression)?'.zip':'';
    return "security_report_dump_$report_id.json".$suffix;
}

#-------------------------------------------------------------------------------

# _report_store_attachment_csv($$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance
#  Comments  :
#  See Also  :

sub _report_store_attachment_csv($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($id, $report) = @_;

    Log::Loger->info(__PACKAGE__, "[ENG:Reporter] Report '$id': Store'n'compressing CSV data attachment");
    my ($plain_f, $zip_f) = _store_and_zip($self->{REPORTS_DIR}, $self->_generate_report_attach_id($id, 'csv'), $report->{attachment_csv});
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter] Report '$id': Store'n'compressed CSV data attachment to files ($plain_f, $zip_f)");

    $report->{flag_archived}++;
}

# _report_store_attachment_json($$$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance
#  Comments  :
#  See Also  :

sub _report_store_attachment_json($$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($id, $report) = @_;

    Log::Loger->info(__PACKAGE__, "[ENG:Reporter] Report '$id': Store'n'compressing JSON data attachment");
    my ($plain_f, $zip_f) = _store_and_zip($self->{REPORTS_DIR}, $self->_generate_report_attach_id($id, 'json'), $report->{attachment_json});
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter] Report '$id': Store'n'compressed JSON data attachment to files ($plain_f, $zip_f)");

    $report->{flag_jarchived}++;
}

# _report_store_file($$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance
#  Comments  :
#  See Also  :

sub _report_store_file($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($id, $report) = @_;

    Log::Loger->info(__PACKAGE__, "[ENG:Reporter] Report '$id': Exporting report into JSON file");
    my $plain_f = _store($self->{REPORTS_DIR}, $self->_generate_report_json_id($id), $report);
    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter] Report '$id': Exported report into JSON file '$plain_f'");

    $report->{flag_filed}++;
}

# _report_email($$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance
#  Comments  :
#  See Also  :

sub _report_email($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($id, $report) = @_;

    my $abuse    = $report->{'abuse'};
    my $type     = $report->{'type'};
    my $severity = $report->{'severity'};

    my $mode = $self->_get_group_config($abuse, 'mode');
    unless ($mode eq MODE_BOTH() or $mode eq $report->{'type'}) {
        Log::Loger->warn(__PACKAGE__, "[ENG:Reporter] Report '$id': Email reporting of '$type' reports is suppressed by '$mode' mode, report will not be mailed");
        $report->{flag_mail_suppressed}++;
        return;
    }

    # Check, if the reporting is suppressed for current abuse contact
    my $mute = $self->_get_group_config($abuse, 'mute');
    if ($mute) {
        Log::Loger->warn(__PACKAGE__, "[ENG:Reporter] Report '$id': Email reporting is muted, report will not be mailed");
        $report->{flag_mail_muted}++;
        return;
    }
    my $redirect = $self->_get_group_config($abuse, 'redirect');
    my $email    = $self->_get_group_config($abuse, 'email');
    my $compress = $self->_get_group_config($abuse, 'compress');
    my $attachs  = lc($self->_get_group_config($abuse, 'attach'));

    $report->{'to'}               = $email;
    $report->{'attachment_file1'} = $self->_generate_report_attach_id($id, 'json', $compress) if $attachs eq 'all' or $attachs eq 'json';
    $report->{'attachment_file2'} = $self->_generate_report_attach_id($id, 'csv',  $compress) if $attachs eq 'all' or $attachs eq 'csv';
    my ($rc, $result, $to) = $self->{MAILER}->email_report($report, redirect => $redirect);

    $report->{'mail_ts'}  = time;
    $report->{'mail_to'}  = $to;
    $report->{'mail_res'} = $result;
    $report->{'cfg_attach'} = $attachs;
    if ($rc) { $report->{'flag_mail_sent'}++; }
    else     { $report->{'flag_mail_error'}++; }
}

# _report_store_log($$$$;$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance
#  Comments  :
#  See Also  :

sub _report_store_log($$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($id, $report, $stop) = @_;

    # Add or edit some items in report log
    $report->{ts} = time;
    delete $report->{attachment_json} if exists $report->{attachment_json};
    delete $report->{attachment_csv}  if exists $report->{attachment_csv};

    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter] Report '$id': Storing report log: ".Dumper($report)) if $DEVEL;

    eval {
        #die 'BSONObj size: 18739938 (0x11DF2E2) is invalid. Size must be between 0 and 16793600(16MB) First element: cnt_analyzers: 11 at /usr/local/lib/site_perl/Mentat/Storage/Mongo.pm line 307.';

        my $result = $self->{RSTORAGE}->update({'_id' => $id}, $report, {'upsert' => 1});
        Log::Loger->notice(__PACKAGE__, "[ENG:Reporter] Report '$id': Stored report to MongoDB document '$id'");

        $report->{flag_stored}++;
    };
    if($@) {
        # Just for sure, save it to other variable
        my $error = $@;
        Log::Loger->error(__PACKAGE__, "[ENG:Reporter] Report '$id': Unable to store report to MongoDB: '$error'");

        # WARNING !!! Beware of recursion, the 'STOP' test is NECESSARY !!!
        if ($error =~ /BSONObj size/ and not $stop) {
            $self->_report_store_file($id, $report);
            Log::Loger->notice(__PACKAGE__, "[ENG:Reporter] Report '$id': Storing only report skeleton to database");
            my $string = "Message was too big to be stored within one database\nobject, it was dumped to the file on disk.\n";

            # WARNING !!! Beware of recursion, the 'STOP' as last argument is NECESSARY !!!
            $self->_report_store_log($id, $string, $string, $report, 'STOP');
        }
        else {
            Log::Loger->error(__PACKAGE__, "[ENG:Reporter] Report '$id': Unable to store report to MongoDB: ".Dumper($report));
        }
    }
}

# _runlog_update($$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on object instance
#  Comments  :
#  See Also  :

sub _runlog_update($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($runlog, $report, $id) = @_;

    $runlog->{reports_all}++;
    $runlog->{"reports_t_".$report->{type}}++;
    $runlog->{"reports_s_".$report->{severity}}++ if $report->{severity};

    $runlog->{reports_a_archived}++  if $report->{flag_archived};
    $runlog->{reports_a_jarchived}++ if $report->{flag_jarchived};
    $runlog->{reports_a_filed}++     if $report->{flag_filed};
    $runlog->{reports_a_mailed}++    if $report->{flag_mail_sent};
    $runlog->{reports_a_stored}++    if $report->{flag_stored};

    map { $runlog->{reports_abuses}->{$_}++ } $self->escape_keys(split(/, /, $report->{abuse}));
    if ($report->{mail_to}) {
        map { $runlog->{reports_mails}->{$_}++ }  $self->escape_keys(split(/, /, $report->{mail_to}));
    }

    push(@{$runlog->{ids}}, $id);

    my ($key, $value);
    while (($key, $value) = each %{$report->{analyzers}}) {
        $runlog->{analyzers}->{$key} += $value;
    }
    while (($key, $value) = each %{$report->{categories}}) {
        $runlog->{categories}->{$key} += $value;
    }

    #print STDERR "RL UPDATE: ".Dumper($runlog, $report);
}

#-------------------------------------------------------------------------------

# _store_and_zip($$$) [PROTECTED]
#
#  Usage     : my ($plain_f, $zip_f) = _store_and_zip($dir, $file_name, $contents);
#  Purpose   : Store given contents into file with given name and compress it
#  Arguments : STRING $dir - target directory
#              STRING $file_name - name of the target file
#              STRING $contents - file contents
#  Returns   : STRING $plain_f - name of the plain text file
#              STRING $zip_f - name of the compressed file
#  Throws    : Dies on error

sub _store_and_zip($$$)
{
    my ($dir, $file_name, $contents) = @_;

    if ($dir =~ /(.*)/) { $dir = $1; }
    chdir $dir;
    my $plain_f = "$file_name";     if ($plain_f =~ /(.*)/) { $plain_f = $1; }
    my $zip_f   = "$file_name.zip"; if ($zip_f =~ /(.*)/)   { $zip_f = $1; }

    # Store the contents to the temporary file
    open(FILE, ">$plain_f") or die "Can't store contents into temporary file '$plain_f': $!\n";
    print FILE $contents;
    close FILE;

    # Zip the file
    zip "$plain_f" => "$zip_f" or die "Compression of file '$plain_f' failed: $ZipError\n";

    # Cleanup
    #unlink "$plain_f";

    # Return result
    return ($plain_f, $zip_f);
}

# _store($$$) [PROTECTED]
#
#  Usage     : my $plain_f = _store($dir, $file_name, $contents);
#  Purpose   : Store given contents into file with given name and compress it
#  Arguments : STRING $dir - target directory
#              STRING $file_name - name of the target file
#              STRING $contents - file contents
#  Returns   : STRING $plain_f - name of the plain text file
#  Throws    : Dies on error

sub _store($$$)
{
    my ($dir, $file_name, $contents) = @_;

    if ($dir =~ /(.*)/) { $dir = $1; }
    chdir $dir;
    my $plain_f = "$file_name";     if ($plain_f =~ /(.*)/) { $plain_f = $1; }

    # Export object to JSON
    my $json = to_json($contents, {utf8 => 1, pretty => 1, allow_blessed => 1, convert_blessed => 1});

    # Store the contents to the temporary file
    open(FILE, ">$plain_f") or die "Can't store object into temporary file '$plain_f': $!\n";
    print FILE $contents;
    close FILE;

    # Return result
    return $plain_f;
}

=pod

=back

=cut

1;
