package Mentat::Stats::Mongo;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Stats::Mongo - Short module description

=head1 SYNOPSIS

 use Mentat::Stats::Mongo;

=head1 DESCRIPTION



=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Mentat::Stats(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Mentat::Stats;
use Mentat::Storage::Mongo;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Mentat::Stats');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance

=cut

sub DESTROY()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Destroying Mentat::Stats::Mongo instance...
}

#-------------------------------------------------------------------------------

=item analyze($;@) [PUBLIC]

  Usage     : my $stats = $mongostats->analyze($database, @collections);
  Purpose   : Analyze MongoDB statistics
  Arguments : STRING $database - name of the database to analyze [MANDATORY]
  Returns   : HASH REFERENCE $stats
  Throws    : Dies, if not invoked on instance

=cut

sub analyze($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($database, @collections) = @_;

    my $dbengine = Mentat::Storage::Mongo->new(database => $database, collection => 'admin');
    my $stats = {};

    @collections = $dbengine->list_collections() unless @collections;
    my $tmp;
    foreach my $collection (@collections) {
        # Attempt to load statistics for collection and ignore any errors
        eval {
            $dbengine->change_collection($collection);
            $stats->{$collection}->{'stats'}   = $dbengine->run_command({'collStats' => $collection});
            $stats->{$collection}->{'indexes'} = $self->_process_indexes([$dbengine->list_indexes()]);
            $stats->{$collection}->{'newest'}  = $self->_process_newold($dbengine->find(undef, undef, {'ts'=>-1}, 1));
            $stats->{$collection}->{'oldest'}  = $self->_process_newold($dbengine->find(undef, undef, {'ts'=>1},  1));
        };
    }

    return $stats;
}

=item analyze_alerts($) [PUBLIC]

  Usage     : my $stats = $mongostats->analyze_alerts($database);
  Purpose   : Analyze alerts collection
  Arguments : STRING $database - name of the database to analyze [MANDATORY]
  Returns   : HASH REFERENCE $stats
  Throws    : Dies, if not invoked on instance

=cut

sub analyze_alerts($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($database,) = @_;

    my $dbengine = Mentat::Storage::Mongo->new(database => $database, collection => 'alerts');
    my $stats = {};

    my $ts    = time;
    my $tsy   = $ts - 31556926;
    my $tstqy = $ts - ((31556926/4)*3);
    my $tshy  = $ts - (31556926/2);
    my $tsty  = $ts - (31556926/3);
    my $tsqy  = $ts - (31556926/4);
    my $tssy  = $ts - (31556926/6);
    my $tsdy  = $ts - (31556926/12);

    $stats->{cnty}   = $dbengine->count({'_CESNET.StorageTime' => {'$lt' => $tsy}});
    $stats->{cnttqy} = $dbengine->count({'_CESNET.StorageTime' => {'$lt' => $tstqy}});
    $stats->{cnthy}  = $dbengine->count({'_CESNET.StorageTime' => {'$lt' => $tshy}});
    $stats->{cntty}  = $dbengine->count({'_CESNET.StorageTime' => {'$lt' => $tsty}});
    $stats->{cntqy}  = $dbengine->count({'_CESNET.StorageTime' => {'$lt' => $tsqy}});
    $stats->{cntsy}  = $dbengine->count({'_CESNET.StorageTime' => {'$lt' => $tssy}});
    $stats->{cntdy}  = $dbengine->count({'_CESNET.StorageTime' => {'$lt' => $tsdy}});

    return $stats;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Stats::Mongo instance
#  Arguments : Unknown see the description in appropriate subclass
#  Returns   : Mentat::Stats::Mongo $self reference
#  Throws    : Dies, if not invoked on instance or given invalid arguments
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Initializing Mentat::Stats::Mongo instance...
    return $self;
}

sub _process_indexes($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($data,) = @_;

    return [sort { $a->{name} cmp $b->{name} } @$data];
}

sub _process_newold($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($data, $count) = @_;

    my $result = {};
    my $item = shift $data;
    return undef unless $item->{ts};

    $result->{'id'} = $item->{'_id'};
    $result->{'ts'} = $item->{'ts'};

    if ($item->{Format} and $item->{Format} =~ /IDEA/) {
        $result->{'ts'} = Value::Convertor->bin_to_bigint($result->{'ts'});
        $result->{'ts'} = Value::Convertor->nbi_to_unixs($result->{'ts'});
        $result->{'ts'} =~ s/\.[0-9]+$//;
    }
    return $result;
}

=pod

=back

=cut

1;
